using System;
using NUnit.Framework;
using SharpSword.Formatters;

namespace SharpSword.Tests
{
   [TestFixture]
   public class HtmlFormatterTests
   {
      IFormatter _formatter;
      IVerseIdentifier _verseIdentifier;
      IVerseIdentifier _verseIdentifierThree;
      
      
      [SetUp]
      public void Setup()
      {
         _formatter = new HtmlFormatter();
         _formatter.IncludeVerseNumbers = false;
         _verseIdentifier = new VerseIdentifier("Genesis",1,1);
         _verseIdentifierThree = new VerseIdentifier("Genesis",1,3);
      }
      
      [Test]
      public void Format_UnknownTags_Removed ()
      {
         //arrange
         var osisText = "text <gas>that</gas>is not removed.";
         var expectedFormattedText = "text is not removed.";

         //Act
         var formattedText = _formatter.Format(osisText, _verseIdentifier);

         //Assert
         Assert.That (formattedText, Contains.Substring(expectedFormattedText));
      }

      [Test]
      public void Format_QuoteTagWithoutMarker_UsesDefaultQuoteMarks()
      {
         //arrange
         var osisText = "text before quote <q>text in quotes</q> text after quote";
         var expectedFormattedText = "text before quote \"text in quotes\" text after quote";

         //Act
         var formattedText = _formatter.Format(osisText, _verseIdentifier);

         //Assert
         Assert.That (formattedText, Contains.Substring(expectedFormattedText));
      }

      [Test]
      public void Format_QuoteTagWithMarker_UsesSpecifiedQuoteMarks()
      {
         //arrange
         var osisText = "text before quote <q marker=\"'\">text in quotes</q> text after quote";
         var expectedFormattedText = "text before quote \'text in quotes\' text after quote";

         //Act
         var formattedText = _formatter.Format(osisText, _verseIdentifier);

         //Assert
         Assert.That (formattedText, Contains.Substring(expectedFormattedText));
      }

      [Test]
      public void Format_QuoteTagStartMilestone_UsesSpecifiedQuoteMarks()
      {
         //arrange
         var osisText = "text before quote <q sID=\"1\" />text in quote...";
         var expectedFormattedText = "text before quote \"text in quote...";;

         //Act
         var formattedText = _formatter.Format(osisText, _verseIdentifier);

         //Assert
         Assert.That (formattedText, Contains.Substring(expectedFormattedText));
      }

      [Test]
      public void Format_QuoteTagEndMilestone_UsesSpecifiedQuoteMarks()
      {
         //arrange
         var osisText = "text in quote<q eID=\"1\" /> text after quote";
         var expectedFormattedText = "text in quote\" text after quote";

         //Act
         var formattedText = _formatter.Format(osisText, _verseIdentifier);

         //Assert
         Assert.That (formattedText, Contains.Substring(expectedFormattedText));
      }

      [Test]
      public void Format_MilestoneCQuoteNoMarker_UsesDefaultQuoteMark()
      {
         //arrange
         var osisText = "text before quote <milestone type=\"cQuote\"/>text after quote";
         var expectedFormattedText = "text before quote \"text after quote";

         //Act
         var formattedText = _formatter.Format(osisText, _verseIdentifier);

         //Assert
         Assert.That (formattedText, Contains.Substring(expectedFormattedText));
      }

      [Test]
      public void Format_MilestoneCQuoteWithMarker_UsesSpecifiedQuoteMarker()
      {
         //arrange
         var osisText = "text before quote <milestone marker=\"“\" type=\"cQuote\"/>text after quote";
         var expectedFormattedText = "text before quote “text after quote";

         //Act
         var formattedText = _formatter.Format(osisText, _verseIdentifier);

         //Assert
         Assert.That (formattedText, Contains.Substring(expectedFormattedText));
      }

      [Test]
      public void Format_LineNoLevelSpecified_EnterAtEndOfline()
      {
         //arrange
         var osisText = @"<l>this is a line</l>";
         var expectedFormattedText = "<div class='firstLine'>this is a line</div>";

         //Act
         var formattedText = _formatter.Format(osisText, _verseIdentifier);

         //Assert
         Assert.That (formattedText, Contains.Substring(expectedFormattedText));
      }

      [Test]
      public void Format_LineMilestoneTypeFirstLine_ChangedIntoFirstLineDiv()
      {
         //arrange
         var osisText = "<l sID=\"34\" />this is a line<l eID=\"34\" />";
         var expectedFormattedText = "<div class='firstLine'>this is a line</div>";

         //Act
         var formattedText = _formatter.Format(osisText, _verseIdentifier);

         //Assert
         Assert.That (formattedText, Contains.Substring(expectedFormattedText));
      }
         
      [Test]
      public void Format_LineMilestoneTypeSecondLine_ChangedIntoLineDiv()
      {
         //arrange
         var osisText = "<l sID=\"34\" />this is the first line<l eID=\"34\" /><l sID=\"34\" />this is the second line<l eID=\"34\" />";
         var expectedFormattedText = "<div class='line'>this is the second line</div>";

         //Act
         var formattedText = _formatter.Format(osisText, _verseIdentifier);

         //Assert
         Assert.That (formattedText, Contains.Substring(expectedFormattedText));
      }
      
      [Test]
      public void Format_FirstLineVerseNumberOn_VerseNumberAddedInsideFirstLine()
      {
         //arrange
         var osisText = "<l sID=\"34\" />this is a line<l eID=\"34\" />";
         var expectedFormattedText = "<div class='firstLine'><div class=\"verseNumber\">1</div> this is a line</div>";
         _formatter.IncludeVerseNumbers = true;
         //Act
         var formattedText = _formatter.Format(osisText, _verseIdentifier);

         //Assert
         Assert.That (formattedText, Contains.Substring(expectedFormattedText));
      }
      
      [Test]
      public void Format_LineGroupMilestoneType_BreakAtStartNotEnd()
      {
         //arrange
         var osisText = "<lg sId=\"32\" /><l sID=\"34\" />this is then end of a line group<l eID=\"34\" /><lg eID=\"32\" />";
         var expectedFormattedText = "<br /><div class='firstLine'>this is then end of a line group</div>";

         //Act
         var formattedText = _formatter.Format(osisText, _verseIdentifier);

         //Assert
         Assert.That (formattedText, Contains.Substring(expectedFormattedText));
      }

      [Test]
      public void Format_DevineName_Included()
      {
         //arrange
         var osisText = "the <divineName>LORD</divineName> said";
         var expectedFormattedText = "the LORD said";

         //Act
         var formattedText = _formatter.Format(osisText, _verseIdentifier);

         //Assert
         Assert.That (formattedText, Contains.Substring(expectedFormattedText));
      }

      [Test]
      public void Format_LineBreak_EnterInserted()
      {
         //arrange
         var osisText = @"this is the first line <lb /> this is the second";
         var expectedFormattedText = "this is the first line <br /> this is the second";

         //Act
         var formattedText = _formatter.Format(osisText, _verseIdentifier);

         //Assert
         Assert.That (formattedText, Contains.Substring(expectedFormattedText));
      }
      
      [Test]
      public void Format_Title_ChangedToHeadingDiv()
      {
         //arrange
         var osisText = "<title subType=\"x-preverse\" type=\"section\">The Beginning of Creation</title>this is the first line";
         var expectedFormattedText = "<div id=\"heading\">The Beginning of Creation</div>";

         //Act
         var formattedText = _formatter.Format(osisText, _verseIdentifier);

         //Assert
         Assert.That (formattedText, Is.StringStarting(expectedFormattedText));
      }
      
      [Test]
      public void Format_VerseNumbersOnNotTitle_VerseAtBeginning()
      {
         //arrange
         var osisText = "This is the third verse.";
         var expectedFormattedText = "<div class=\"verseNumber\">3</div>This is the third verse.";
         _formatter.IncludeVerseNumbers = true;
         //Act
         var formattedText = _formatter.Format(osisText, _verseIdentifierThree);

         //Assert
         Assert.That (formattedText, Contains.Substring(expectedFormattedText));
      }
      
      [Test]
      public void Format_VerseNumbersOnVerseStartsWithTitle_TitleComesFirstThenAnchorThenVerseDivThenNumber()
      {
         //arrange
         var osisText = "<title>Some Title</title>This is the third verse.";
         var expectedFormattedText = "<div id=\"heading\">Some Title</div><a name=\"Genesis 1:3\"></a><div class=\"verse\"><div class=\"verseNumber\">3</div>";
         _formatter.IncludeVerseNumbers = true;
         //Act
         var formattedText = _formatter.Format(osisText, _verseIdentifierThree);

         //Assert
         Assert.That (formattedText, Is.StringStarting(expectedFormattedText));
      }
      
      [Test]
      public void Format_NoTitle_HasAnchorTag()
      {
         //arrange
         var osisText = "verse text";
         var expectedFormattedText = "<a name=\"Genesis 1:1\"></a>";
         
         //Act
         var formattedText = _formatter.Format(osisText, _verseIdentifier);

         //Assert
         StringAssert.StartsWith(expectedFormattedText, formattedText);
      }
      
      [Test]
      public void Format_NoTitle_HasVerseDivTag()
      {
         //arrange
         var osisText = "verse text";
         var expectedFormattedText = "<div class=\"verse\">verse text</div>";
         
         //Act
         var formattedText = _formatter.Format(osisText, _verseIdentifier);

         //Assert
         Assert.That (formattedText, Contains.Substring(expectedFormattedText));
      }
      
      [Test]
      public void Format_NoteA_ConvertedToNoteLink()
      {
         //arrange
         var osisText = "<note n=\"a\">note contents</note>";
         var expectedFormattedText 
            = "<a href=\"#\" class='notelink'>a</a><div class=\"note\">note contents</div>";
         
         //Act
         var formattedText = _formatter.Format(osisText, _verseIdentifier);

         //Assert
         Assert.That (formattedText, Contains.Substring(expectedFormattedText));
      }
      
      [Test]
      public void Format_NoteB_ConvertedToNoteLink()
      {
         //arrange
         var osisText = "<note n=\"b\">note contents</note>";
         var expectedFormattedText 
            = "<a href=\"#\" class='notelink'>b</a><div class=\"note\">note contents</div>";
         
         //Act
         var formattedText = _formatter.Format(osisText, _verseIdentifier);

         //Assert
         Assert.That (formattedText, Contains.Substring(expectedFormattedText));
      }
      
      [Test]
      public void Format_Reference_ConvertedToLink()
      {
         //arrange
         var osisText = "<reference osisRef=\"Prov.10.1\">ch. 10:1</reference>";
         var expectedFormattedText 
            = "<a href=\"Proverbs 10:1\">ch. 10:1</a>";
         
         //Act
         var formattedText = _formatter.Format(osisText, _verseIdentifier);

         //Assert
         Assert.That (formattedText, Contains.Substring(expectedFormattedText));
      }
      
      [Test]
      public void Format_ReferenceNoVerseSpecified_ConvertedToLinkPointingAtFirstVerse()
      {
         //arrange
         var osisText = "<reference osisRef=\"Prov.10\">ch. 10:1</reference>";
         var expectedFormattedText 
            = "<a href=\"Proverbs 10:1\">ch. 10:1</a>";
         
         //Act
         var formattedText = _formatter.Format(osisText, _verseIdentifier);

         //Assert
         Assert.That (formattedText, Contains.Substring(expectedFormattedText));
      }
      
      //Num.9.17-Num.9.23
      [Test]
      public void Format_OsisReferenceContainsDash_ConvertedToLink()
      {
         //arrange
         var osisText = "<reference osisRef=\"Num.9.17-Num.9.23\">ch. 10:1</reference>";
         var expectedFormattedText 
            = "<a href=\"Numbers 9:17\">ch. 10:1</a>";
         
         //Act
         var formattedText = _formatter.Format(osisText, _verseIdentifier);

         //Assert
         Assert.That (formattedText, Contains.Substring(expectedFormattedText));
      }
      
      [Test]
      public void Format_HiNoType_ContentsIncluded()
      {
         //arrange
         var osisText = "before <hi>contents</hi> after";
         var expectedFormattedText 
            = "before contents after";
         
         //Act
         var formattedText = _formatter.Format(osisText, _verseIdentifier);

         //Assert
         Assert.That (formattedText, Contains.Substring(expectedFormattedText));
      }
      
      [Test]
      public void Format_HiNormalType_ContentsIncluded()
      {
         //arrange
         var osisText = "before <hi type=\"normal\">contents</hi> after";
         var expectedFormattedText 
            = "before contents after";
         
         //Act
         var formattedText = _formatter.Format(osisText, _verseIdentifier);

         //Assert
         Assert.That (formattedText, Contains.Substring(expectedFormattedText));
      }
      
      [Test]
      public void Format_HiEmphasisType_ContentsB()
      {
         //arrange
         var osisText = "before <hi type=\"bold\">contents</hi> after";
         var expectedFormattedText 
            = "before <b>contents</b> after";
         
         //Act
         var formattedText = _formatter.Format(osisText, _verseIdentifier);

         //Assert
         Assert.That (formattedText, Contains.Substring(expectedFormattedText));
      }
      
      [Test]
      public void Format_HiEmphasisType_ContentsEm()
      {
         //arrange
         var osisText = "before <hi type=\"emphasis\">contents</hi> after";
         var expectedFormattedText 
            = "before <em>contents</em> after";
         
         //Act
         var formattedText = _formatter.Format(osisText, _verseIdentifier);

         //Assert
         Assert.That (formattedText, Contains.Substring(expectedFormattedText));
      }
      
      [Test]
      public void Format_HiItalicType_ContentsI()
      {
         //arrange
         var osisText = "before <hi type=\"italic\">contents</hi> after";
         var expectedFormattedText 
            = "before <i>contents</i> after";
         
         //Act
         var formattedText = _formatter.Format(osisText, _verseIdentifier);

         //Assert
         Assert.That (formattedText, Contains.Substring(expectedFormattedText));
      }
      
      [Test]
      public void Format_HiLineThroughType_ContentsStrike()
      {
         //arrange
         var osisText = "before <hi type=\"line-through\">contents</hi> after";
         var expectedFormattedText 
            = "before <strike>contents</strike> after";
         
         //Act
         var formattedText = _formatter.Format(osisText, _verseIdentifier);

         //Assert
         Assert.That (formattedText, Contains.Substring(expectedFormattedText));
      }
      
      [Test]
      public void Format_HiSmallCapsType_ContentsSmallCaps()
      {
         //arrange
         var osisText = "before <hi type=\"small-caps\">contents</hi> after";
         var expectedFormattedText 
            = "before <div style=\"font-variant:small-caps; display: inline;\">contents</div> after";
         
         //Act
         var formattedText = _formatter.Format(osisText, _verseIdentifier);

         //Assert
         Assert.That (formattedText, Contains.Substring(expectedFormattedText));
      }
      
      [Test]
      public void Format_HiSubType_ContentsSub()
      {
         //arrange
         var osisText = "before <hi type=\"sub\">contents</hi> after";
         var expectedFormattedText 
            = "before <sub>contents</sub> after";
         
         //Act
         var formattedText = _formatter.Format(osisText, _verseIdentifier);

         //Assert
         Assert.That (formattedText, Contains.Substring(expectedFormattedText));
      }
      
      [Test]
      public void Format_HiSuperType_ContentsSup()
      {
         //arrange
         var osisText = "before <hi type=\"super\">contents</hi> after";
         var expectedFormattedText 
            = "before <sup>contents</sup> after";
         
         //Act
         var formattedText = _formatter.Format(osisText, _verseIdentifier);

         //Assert
         Assert.That (formattedText, Contains.Substring(expectedFormattedText));
      }
      
      [Test]
      public void Format_HiUnderlineType_ContentsU()
      {
         //arrange
         var osisText = "before <hi type=\"underline\">contents</hi> after";
         var expectedFormattedText 
            = "before <u>contents</u> after";
         
         //Act
         var formattedText = _formatter.Format(osisText, _verseIdentifier);

         //Assert
         Assert.That (formattedText, Contains.Substring(expectedFormattedText));
      }
      
      [Test]
      public void ChapterTitle_Genesis1_CorrectText()
      {
         //arrange
         var genesisOne = new VerseIdentifier("Genesis", 1, 1);
         
         //act
         var chapterTitle = _formatter.ChapterTitle(genesisOne);
         
         //assert
         Assert.That (
            chapterTitle, 
            Is.EqualTo ("<h1>Chapter 1</h1>"));
      }
      
      [Test]
      public void ChapterTitle_Genesis21_CorrectText()
      {
         //arrange
         var genesisOne = new VerseIdentifier("Genesis", 21, 1);
         
         //act
         var chapterTitle = _formatter.ChapterTitle(genesisOne);
         
         //assert
         Assert.That (
            chapterTitle, 
            Is.EqualTo ("<h1>Chapter 21</h1>"));
      }    
   }
}

