using System;
using NUnit.Framework;
using SharpSword.Formatters;

namespace SharpSword.Tests
{
   [TestFixture]
   public class PlainTests
   {
      IVerseIdentifier _verseIdentifier;
      IFormatter _plainFormatter;
      
      [SetUp]
      public void Setup()
      {
         _verseIdentifier = new VerseIdentifier("Genesis",1,1);
         _plainFormatter = new PlainFormatter();
      }
      
      [Test]
      public void PlainTest_FormatGen1v1_RemovesTags ()
      {
         //arrange
         var osisText = "<title subType=\"x-preverse\" type=\"section\">The Beginning of Creation</title>In the <note n=\"a\" osisID=\"Gen.1.1!crossReference.a\" osisRef=\"Gen.1.1\" type=\"crossReference\"><reference osisRef=\"Job.38.4-Job.38.7\">Job 38:4-7</reference>; <reference osisRef=\"Ps.33.6\">Ps. 33:6</reference>; <reference osisRef=\"Ps.136.5\">136:5</reference>; <reference osisRef=\"Isa.42.5\">Isa. 42:5</reference>; <reference osisRef=\"Isa.45.18\">45:18</reference>; <reference osisRef=\"John.1.1-John.1.3\">John 1:1-3</reference>; <reference osisRef=\"Acts.14.15\">Acts 14:15</reference>; <reference osisRef=\"Acts.17.24\">17:24</reference>; <reference osisRef=\"Col.1.16-Col.1.17\">Col. 1:16, 17</reference>; <reference osisRef=\"Heb.1.10\">Heb. 1:10</reference>; <reference osisRef=\"Heb.11.3\">11:3</reference>; <reference osisRef=\"Rev.4.11\">Rev. 4:11</reference></note>beginning, God created the heavens and the earth.";
         var expectedFormattedText = "In the beginning, God created the heavens and the earth.";
         var plain = new PlainFormatter();

         //Act
         var formattedText = plain.Format(osisText, _verseIdentifier);

         //Assert
         Assert.That (formattedText, Is.EqualTo (expectedFormattedText));
      }

      [Test]
      public void Format_QuoteTagWithoutMarker_UsesDefaultQuoteMarks()
      {
         //arrange
         var osisText = "text before quote <q>text in quotes</q> text after quote";
         var expectedFormattedText = "text before quote \"text in quotes\" text after quote";
         var plain = new PlainFormatter();

         //Act
         var formattedText = plain.Format(osisText, _verseIdentifier);

         //Assert
         Assert.That (formattedText, Is.EqualTo (expectedFormattedText));
      }

      [Test]
      public void Format_QuoteTagWithMarker_UsesSpecifiedQuoteMarks()
      {
         //arrange
         var osisText = "text before quote <q marker=\"'\">text in quotes</q> text after quote";
         var expectedFormattedText = "text before quote \'text in quotes\' text after quote";
         var plain = new PlainFormatter();

         //Act
         var formattedText = plain.Format(osisText, _verseIdentifier);

         //Assert
         Assert.That (formattedText, Is.EqualTo (expectedFormattedText));
      }

      [Test]
      public void Format_QuoteTagStartMilestone_UsesSpecifiedQuoteMarks()
      {
         //arrange
         var osisText = "text before quote <q sID=\"1\" />text in quote...";
         var expectedFormattedText = "text before quote \"text in quote...";
         var plain = new PlainFormatter();

         //Act
         var formattedText = plain.Format(osisText, _verseIdentifier);

         //Assert
         Assert.That (formattedText, Is.EqualTo (expectedFormattedText));
      }

      [Test]
      public void Format_QuoteTagEndMilestone_UsesSpecifiedQuoteMarks()
      {
         //arrange
         var osisText = "text in quote<q eID=\"1\" /> text after quote";
         var expectedFormattedText = "text in quote\" text after quote";
         var plain = new PlainFormatter();

         //Act
         var formattedText = plain.Format(osisText, _verseIdentifier);

         //Assert
         Assert.That (formattedText, Is.EqualTo (expectedFormattedText));
      }

      [Test]
      public void Format_MilestoneCQuoteNoMarker_UsesDefaultQuoteMark()
      {
         //arrange
         var osisText = "text before quote <milestone type=\"cQuote\"/>text after quote";
         var expectedFormattedText = "text before quote \"text after quote";
         var plain = new PlainFormatter();

         //Act
         var formattedText = plain.Format(osisText, _verseIdentifier);

         //Assert
         Assert.That (formattedText, Is.EqualTo (expectedFormattedText));
      }

      [Test]
      public void Format_MilestoneCQuoteWithMarker_UsesSpecifiedQuoteMarker()
      {
         //arrange
         var osisText = "text before quote <milestone marker=\"“\" type=\"cQuote\"/>text after quote";
         var expectedFormattedText = "text before quote “text after quote";
         var plain = new PlainFormatter();

         //Act
         var formattedText = plain.Format(osisText, _verseIdentifier);

         //Assert
         Assert.That (formattedText, Is.EqualTo (expectedFormattedText));
      }

      [Test]
      public void Format_LineNoLevelSpecified_EnterAtEndOfline()
      {
         //arrange
         var osisText = @"<l>this is a line</l>";
         var expectedFormattedText = "this is a line" + Environment.NewLine;
         var plain = new PlainFormatter();

         //Act
         var formattedText = plain.Format(osisText, _verseIdentifier);

         //Assert
         Assert.That (formattedText, Is.EqualTo (expectedFormattedText));
      }

      [Test]
      public void Format_LineMilestoneType_EnterAtEndOfline()
      {
         //arrange
         var osisText = "<l sID=\"34\" />this is a line<l eID=\"34\" />";
         var expectedFormattedText = "this is a line" + Environment.NewLine;
         var plain = new PlainFormatter();

         //Act
         var formattedText = plain.Format(osisText, _verseIdentifier);

         //Assert
         Assert.That (formattedText, Is.EqualTo (expectedFormattedText));
      }

      [Test]
      public void Format_DevineName_Included()
      {
         //arrange
         var osisText = "the <divineName>LORD</divineName> said";
         var expectedFormattedText = "the LORD said";
         var plain = new PlainFormatter();

         //Act
         var formattedText = plain.Format(osisText, _verseIdentifier);

         //Assert
         Assert.That (formattedText, Is.EqualTo (expectedFormattedText));
      }

      [Test]
      public void Format_LineBreak_EnterInserted()
      {
         //arrange
         var osisText = @"this is the first line <lb /> this is the second";
         var expectedFormattedText = "this is the first line " + Environment.NewLine + " this is the second";
         var plain = new PlainFormatter();

         //Act
         var formattedText = plain.Format(osisText, _verseIdentifier);

         //Assert
         Assert.That (formattedText, Is.EqualTo (expectedFormattedText));
      }

      [Test]
      public void Format_ContainsNewLineCharacter_Removed()
      {
         //arrange
         var osisText = "before new line char \nafter new line char";
         var expectedFormattedText = "before new line char after new line char";
         var plain = new PlainFormatter();

         //Act
         var formattedText = plain.Format(osisText, _verseIdentifier);

         //Assert
         Assert.That (formattedText, Is.EqualTo (expectedFormattedText));
      }

      [Test]
      public void Format_ContainsExtraSpaces_Removed()
      {
         //arrange
         var osisText = "one two  three   four    ";
         var expectedFormattedText = "one two three four ";
         var plain = new PlainFormatter();

         //Act
         var formattedText = plain.Format(osisText, _verseIdentifier);

         //Assert
         Assert.That (formattedText, Is.EqualTo (expectedFormattedText));
      }
      
      [Test]
      public void ChapterTitle_Genesis1_CorrectText()
      {
         //arrange
         var genesisOne = new VerseIdentifier("Genesis", 1, 1);
         
         //act
         var chapterTitle = _plainFormatter.ChapterTitle(genesisOne);
         
         //assert
         Assert.That (
            chapterTitle, 
            Is.EqualTo (string.Format ("Chapter 1{0}", Environment.NewLine)));
      }
      
      [Test]
      public void ChapterTitle_Genesis21_CorrectText()
      {
         //arrange
         var genesisOne = new VerseIdentifier("Genesis", 21, 1);
         
         //act
         var chapterTitle = _plainFormatter.ChapterTitle(genesisOne);
         
         //assert
         Assert.That (
            chapterTitle, 
            Is.EqualTo (string.Format ("Chapter 21{0}", Environment.NewLine)));
      }
   }
}

