using System;
using NUnit.Framework;
using Moq;
using System.IO;
using System.Collections.Generic;
using System.Linq;

namespace SharpSword.Tests
{
   [TestFixture]
   public class ModuleManagerTests
   {
      static string SWORD_APP_DATA = ModuleManager.GetSwordAppDataPath();
      static string SWORD_MODD_PATH = Path.Combine (SWORD_APP_DATA, "mods.d");

      static string KJV_PATH = Path.Combine(SWORD_APP_DATA, "modules", "texts", "ztext", "kjv");

      FuncMock<string, IModuleConfig> _moduleConfigFactoryMock;
      Mock<IDirectory> _directoryMock;
      ModuleManager _moduleManager;
      FuncMock<string, IBibleModule> _bibleModuleFactoryMock;
      Mock<IFtp> _ftpMock;
      Mock<IFile> _fileMock;
      Mock<ILuceneIndex> _luceneIndexMock;

      [SetUp]
      public void Setup()
      {
         _moduleConfigFactoryMock = new FuncMock<string, IModuleConfig>();
         _bibleModuleFactoryMock = new FuncMock<string, IBibleModule>();

         _directoryMock = new Mock<IDirectory>();
         _directoryMock
            .Setup (directory => directory.Exists(Path.Combine(KJV_PATH, "lucenenet")))
            .Returns(true);

         _ftpMock = new Mock<IFtp>();
         _fileMock = new Mock<IFile>();

         _luceneIndexMock = new Mock<ILuceneIndex>();
         Func<string, IVersificationInternal, ILuceneIndex> luceneFactory = (path, versification) =>
         {
            Assert.That (path, Is.EqualTo(KJV_PATH));
            return _luceneIndexMock.Object;
         };

         _moduleManager = new ModuleManager (
            luceneFactory,
            () => _ftpMock.Object,
            _moduleConfigFactoryMock.Object,
            _directoryMock.Object,
            _bibleModuleFactoryMock.Object,
            _fileMock.Object);
      }
  
      [Test]
      public void InstalledModules_TempConfig_NotReturned()
      {
         //arrange
         var tempConfigPath = Path.Combine(SWORD_MODD_PATH, "esv.conf.temp");

         _directoryMock
            .Setup (dir => dir.GetFiles(SWORD_MODD_PATH))
            .Returns (new string[]
            {
               tempConfigPath,
            });

         var tempModuleConfigMock = new Mock<IModuleConfig>();

         _moduleConfigFactoryMock.Add (tempConfigPath, tempModuleConfigMock.Object);

         //Act
         var installedModules = _moduleManager.InstalledModules();

         //Assert
         Assert.That (installedModules.Contains (tempModuleConfigMock.Object), Is.False);
      }
      
      [Test]
      public void InstalledModules_TwoInstalled_TwoReturned()
      {
         //arrange
         var kjvConfigPath = Path.Combine(SWORD_MODD_PATH, "kjv.conf");
         var esvConfigPath = Path.Combine(SWORD_MODD_PATH, "esv.conf");

         _directoryMock
            .Setup (dir => dir.GetFiles(SWORD_MODD_PATH))
            .Returns (new string[]
            {
               kjvConfigPath,
               esvConfigPath,
            });

         var kjvModuleConfigMock = new Mock<IModuleConfig>();
         var esvModuleConfigMock = new Mock<IModuleConfig>();

         _moduleConfigFactoryMock.Add (kjvConfigPath, kjvModuleConfigMock.Object);
         _moduleConfigFactoryMock.Add (esvConfigPath, esvModuleConfigMock.Object);

         //Act
         var installedModules = _moduleManager.InstalledModules();

         //Assert
         Assert.That (installedModules.Count(), Is.EqualTo(2));
         Assert.That (installedModules.Contains (kjvModuleConfigMock.Object));
         Assert.That (installedModules.Contains (esvModuleConfigMock.Object));
      }

      [Test]
      public void GetModule_Kjv_ReturnsKjvModule()
      {
         //Arrange
         var bibleModuleMock = new Mock<IBibleModule>();
         _bibleModuleFactoryMock.Add (KJV_PATH, bibleModuleMock.Object);

         var kjvModuleConfigMock = new Mock<IModuleConfig>();
         kjvModuleConfigMock
            .Setup (config => config.DataPath)
            .Returns("./modules/texts/ztext/kjv/");
         kjvModuleConfigMock
            .Setup (config => config.Versification)
            .Returns("kjv");

         var kjvConfigPath = Path.Combine(SWORD_MODD_PATH, "kjv.conf");
         _moduleConfigFactoryMock.Add(kjvConfigPath, kjvModuleConfigMock.Object);

         //Act
         var module = _moduleManager.GetModule("kjv");

         //Assert
         Assert.That (module, Is.EqualTo(bibleModuleMock.Object));
      }

      [Test]
      public void GetModule_IndexDoesntExist_IndexGenerated()
      {
         //Arrange
         var bibleModuleMock = new Mock<IBibleModule>();

         _bibleModuleFactoryMock.Add (KJV_PATH, bibleModuleMock.Object);

         var kjvModuleConfigMock = new Mock<IModuleConfig>();
         kjvModuleConfigMock
            .Setup (config => config.DataPath)
            .Returns("./modules/texts/ztext/kjv/");
         kjvModuleConfigMock
            .Setup (config => config.Versification)
            .Returns("kjv");

         var kjvConfigPath = Path.Combine(SWORD_MODD_PATH, "kjv.conf");
         _moduleConfigFactoryMock.Add(kjvConfigPath, kjvModuleConfigMock.Object);

         _directoryMock
            .Setup (directory => directory.Exists(Path.Combine(KJV_PATH, "lucenenet")))
            .Returns(false);

         var versificationMock = new Mock<IVersificationInternal>();

         var versificationManagerMock = new Mock<IVersificationManager>();
         versificationManagerMock
            .Setup (versificationManager => versificationManager.GetVersification("kjv"))
            .Returns(versificationMock.Object);

         VersificationManager.InjectInstance(versificationManagerMock.Object);

         //Act
         _moduleManager.GetModule("kjv");

         //Assert
         _luceneIndexMock
            .Verify (index => index.GenerateIndex(bibleModuleMock.Object, versificationMock.Object), Times.Once());
      }

      [Test]
      public void GetDownloadableModules_TwoModules_BothReturned()
      {
         //Arrange
         string tempPathOfArchive = null;
         var compressedModsdPath = @"ftp://ftp.crosswire.org/pub/sword/raw/mods.d.tar.gz";

         var esvModuleConfigMock = new Mock<IModuleConfig>();
         var kjvModuleConfigMock = new Mock<IModuleConfig>();

         _ftpMock
            .Setup (ftp => ftp.DownloadFile(compressedModsdPath, It.IsAny<string>()))
            .Callback<string, string>((address, path) =>
            {
               //all the rest of the setup needs to know what temp path was used
               tempPathOfArchive = path;
               var tempPath = Path.GetDirectoryName(tempPathOfArchive);
               var tempModdPath = Path.Combine (tempPath, "mods.d");
               var esvConfigPath = Path.Combine(tempModdPath, "esv.conf");
               var kjvConfigPath = Path.Combine(tempModdPath, "kjv.conf");
               var expectedModules = new string[]
               {
                  esvConfigPath,
                  kjvConfigPath
               };
               _directoryMock
                  .Setup (directory => directory.GetFiles (tempModdPath))
                  .Returns(expectedModules);

               _moduleConfigFactoryMock
                  .Add(esvConfigPath, esvModuleConfigMock.Object)
                  .Add(kjvConfigPath, kjvModuleConfigMock.Object);
            });

         //Act
         var downloadableModules = _moduleManager.GetDownloadableModules();

         //Assert
         Assert.That (downloadableModules.Contains (esvModuleConfigMock.Object));
         Assert.That (downloadableModules.Contains (kjvModuleConfigMock.Object));
      }

      [TearDown]
      public void TearDown()
      {
         VersificationManager.InjectInstance(null);
      }
   }
}

