using System;
using System.Collections.Generic;
using NUnit.Framework;
using SharpSword.reader;
using System.IO;
using System.Linq;
using Moq;

namespace SharpSword.Tests
{
   [TestFixture]
   public class ModuleManager_InstallModuleTests
   {
      Mock<IModuleConfig> _moduleConfigMock;
      Mock<IDirectory> _directoryMock;
      Mock<IFtp> _ftpMock;
      ModuleManager _moduleManager;
      Mock<IVersificationManager> _versificationManagerMock;
      Mock<ILuceneIndex> _luceneIndexMock;
      Mock<IBibleModule> _bibleModuleMock;
      Mock<IVersificationInternal> _versificationMock;
      Mock<IFile> _fileMock;

      [SetUp]
      public void Setup()
      {
         _moduleConfigMock = new Mock<IModuleConfig>();
         _moduleConfigMock
            .Setup (config => config.DataPath)
            .Returns(@"./modules/texts/ztext/kjv/");
         _moduleConfigMock
            .Setup(config => config.Versification)
            .Returns ("kjv");

         _directoryMock = new Mock<IDirectory>();
         _directoryMock
            .Setup (dir => dir.Exists(It.IsAny<string>()))
            .Returns (true);

         _ftpMock = new Mock<IFtp>();


         _versificationMock = new Mock<IVersificationInternal>();

         _versificationManagerMock = new Mock<IVersificationManager>();
         _versificationManagerMock
            .Setup (versificationManager => versificationManager.GetVersification("kjv"))
            .Returns(_versificationMock.Object);

         VersificationManager.InjectInstance(_versificationManagerMock.Object);

         _luceneIndexMock = new Mock<ILuceneIndex>();

         _bibleModuleMock = new Mock<IBibleModule>();
   
         _fileMock = new Mock<IFile>();
         
         _moduleManager = new ModuleManager(
            (module, versification) => _luceneIndexMock.Object,
            () => _ftpMock.Object,
            path => _moduleConfigMock.Object,
            _directoryMock.Object,
            path => _bibleModuleMock.Object,
            _fileMock.Object);
      }

      [Test]
      public void InstallModule_Kjv_DownloadsConfig()
      {
         //arrange
         var expectedPathToDownloadTo = Path.Combine (ModuleManager.GetSwordAppDataPath(), "mods.d", "kjv.conf.tmp");

         //act
         _moduleManager.InstallModule("kjv");

         //assert
         _ftpMock.Verify(
            ftp => ftp.DownloadFile(@"ftp://ftp.crosswire.org/pub/sword/raw/mods.d/kjv.conf", expectedPathToDownloadTo),
            Times.Once());
      }
      
      [Test]
      public void InstallModule_Kjv_RenamesTmpConfigOnceFinished()
      {
         //arrange
         var expectedPathToDownloadTo = Path.Combine (ModuleManager.GetSwordAppDataPath(), "mods.d", "kjv.conf.tmp");
         var expectedPathToRenameTo = Path.Combine (ModuleManager.GetSwordAppDataPath(), "mods.d", "kjv.conf");

         //act
         _moduleManager.InstallModule("kjv");

         //assert
         _fileMock
            .Verify(file => file.Move (expectedPathToDownloadTo, expectedPathToRenameTo));
      }

      [Test]
      public void InstallModule_Kjv_CallsGetFiles()
      {
         //act
         _moduleManager.InstallModule("kjv");

         //assert
         _ftpMock.Verify(ftp => ftp.GetFiles(@"ftp://ftp.crosswire.org/pub/sword/raw/modules/texts/ztext/kjv/"), Times.Once());
      }

      [Test]
      public void InstallModule_Kjv_DownloadsFilesReturnedByGetFiles()
      {
         //arrange
         _ftpMock
            .Setup (ftp => ftp.GetFiles(It.IsAny<string>()))
            .Returns(new string[]{"firstFile", "secondFile"});

         var expectedFolderToDownloadTo =
            Path.Combine (
               ModuleManager.GetSwordAppDataPath(), "modules", "texts", "ztext", "kjv");

         //act
         _moduleManager.InstallModule("kjv");

         //assert
         _ftpMock
            .Verify (
               ftp => ftp.DownloadFile( @"ftp://ftp.crosswire.org/pub/sword/raw/modules/texts/ztext/kjv/firstFile", Path.Combine(expectedFolderToDownloadTo,"firstFile")),
               Times.Once());
         _ftpMock
            .Verify (
               ftp => ftp.DownloadFile( @"ftp://ftp.crosswire.org/pub/sword/raw/modules/texts/ztext/kjv/secondFile", Path.Combine(expectedFolderToDownloadTo,"secondFile")),
               Times.Once());
      }

      [Test]
      public void InstallModule_Kjv_UsesCorrectVersification()
      {
         //act
         _moduleManager.InstallModule("kjv");

         //assert
         _versificationManagerMock
            .Verify(manager => manager.GetVersification("kjv"));
      }

      [Test]
      public void InstallModule_Kjv_GeneratesIndex()
      {
         //act
         _moduleManager.InstallModule("kjv");

         //assert
         _luceneIndexMock
            .Verify(index => index.GenerateIndex(_bibleModuleMock.Object, _versificationMock.Object), Times.Once());
      }

      [TearDown]
      public void TearDown()
      {
         VersificationManager.InjectInstance(null);
      }
   }
}

