
using System;
using System.Xml;
using System.Xml.Linq;
using System.Linq;
using System.Text;
using System.Collections.Generic;
using System.IO;
using System.Reflection;

namespace SharpSword.Formatters
{
   class HtmlFormatter : IFormatter
   {
      static readonly IVersificationInternal _versification = VersificationManager.Instance.GetVersification("kjv");
      
      public HtmlFormatter()
      {
         IncludeVerseNumbers = true;
      }
      
      static string ReadEmbeddedText(string fileName)
      {
         using (Stream stream = Assembly.GetExecutingAssembly()
                                        .GetManifestResourceStream("SharpSword." + fileName))
         using (StreamReader reader = new StreamReader(stream))
         {
             return reader.ReadToEnd();
         }
      }
      
      static readonly string _header = 
         "<!DOCTYPE html>" +
         "<html>" +
         "    <head>" +
         "        <meta http-equiv=\"Content-Type\" content=\"text/html;charset=UTF-8\">" + 
         "        <style>" +
         ReadEmbeddedText("bible.css") +
         "        </style>" +
         "        <script>" +
         ReadEmbeddedText("bible.js") +
         "        </script>" +
         "    </head>" + 
         "    <div id=content>";
      
      public string Header()
      {
         return _header;  
      }
      
      bool _titleFound = false;
      bool _isFirstline = true;
      bool _textAdded = false;
      int _verseNumberIndex = 0;
      
      
      public string Format (string text, IVerseIdentifier verseIdentifier)
      {
         _titleFound = false;
         _isFirstline = true;
         _textAdded = false;
         _verseNumberIndex = 0;
         XElement verse = XElement.Parse("<v>" + text + "</v>", LoadOptions.PreserveWhitespace);
         var stringBuilder = new StringBuilder();
         ParseNode(verse, stringBuilder, verseIdentifier);
         if(IncludeVerseNumbers)
         {
             stringBuilder.Insert(_verseNumberIndex, string.Format("<div class=\"verseNumber\">{0}</div>", verseIdentifier.VerseNumber));
         }
         if(_titleFound == false)
         {
            stringBuilder.Insert (0, "<div class=\"verse\">");
            stringBuilder.Insert (0, string.Format ("<a name=\"{0}\"></a>", verseIdentifier));
         }
         stringBuilder.Append("</div>");
         return stringBuilder.ToString().Replace("<br /> <br /> <br />","<br /> <br />"); //remove double blank lines.
      }
      
      public string Footer()
      {
         return "</div></html>";
      }
      
      public bool IncludeVerseNumbers
      {
         get;
         set;
      }
      
      void ParseNode(XElement element, StringBuilder stringBuilder, IVerseIdentifier verseIdentifier)
      {
         foreach (XNode node in element.Nodes())
         {
            if(node.NodeType == XmlNodeType.Element)
            {
               var elementNode = node as XElement;
               if(elementNode.Name == "q")
               {
                  ProcessQuoteElement(elementNode, stringBuilder, verseIdentifier);
               }
               else if(elementNode.Name == "milestone")
               {
                  ProcessMilestoneElement(elementNode, stringBuilder);
               }
               else if(elementNode.Name == "lg")
               {
                  ProcessLineGroupElement(elementNode, stringBuilder);
               }
               else if(elementNode.Name == "l")
               {
                  ProcessLineElement(elementNode, stringBuilder, verseIdentifier);
               }
               else if(elementNode.Name == "divineName")
               {
                  ParseNode ((XElement)elementNode, stringBuilder, verseIdentifier);
               }
               else if(elementNode.Name == "lb")
               {
                  stringBuilder.Append("<br />");
               }
               else if(elementNode.Name == "title")
               {
                  ProcessTitleElement(elementNode, stringBuilder, verseIdentifier);
               }
               else if(elementNode.Name == "note")
               {
                  ProcessNoteElement(elementNode, stringBuilder, verseIdentifier);
               }
               else if(elementNode.Name == "reference")
               {
                  ProcessReferenceElement(elementNode, stringBuilder, verseIdentifier);
               }
               else if(elementNode.Name == "hi")
               {
                  ProcessHiElement(elementNode, stringBuilder, verseIdentifier);
               }
            }
            else if(node.NodeType == XmlNodeType.Text)
            {
               stringBuilder.Append(node.ToString());
               _textAdded = true;
            }
         }
      }
      
      void ProcessHiElement(XElement elementNode, StringBuilder stringBuilder, IVerseIdentifier verseIdentifier)
      {
         var typeAttribute = elementNode.Attributes("type").FirstOrDefault();
         var type = typeAttribute != null ? typeAttribute.Value : null;
         string opening = string.Empty;
         string closing =  string.Empty;

         if(type == "bold")
         {
            opening = "<b>";
            closing = "</b>";
         }
         else if(type == "emphasis")
         {
            opening = "<em>";
            closing = "</em>";
         }
         else if(type == "italic")
         {
            opening = "<i>";
            closing = "</i>";
         }
         else if(type == "line-through")
         {
            opening = "<strike>";
            closing = "</strike>";
         }
         else if(type == "small-caps")
         {
            opening = "<div style=\"font-variant:small-caps; display: inline;\">";
            closing = "</div>";
         }
         else if(type == "super")
         {
            opening = "<sup>";
            closing = "</sup>";
         }
         else if(type == "sub")
         {
            opening = "<sub>";
            closing = "</sub>";
         }
         else if(type == "underline")
         {
            opening = "<u>";
            closing = "</u>";
         }
         stringBuilder.Append(opening);
         ParseNode ((XElement)elementNode, stringBuilder, verseIdentifier);
         stringBuilder.Append(closing);
      }
      
      void ProcessReferenceElement(XElement elementNode, StringBuilder stringBuilder, IVerseIdentifier verseIdentifier)
      {
         string target = string.Empty;
         var references = elementNode.Attributes("osisRef");
         if(references.Any())
         {
            var targetOsis = references.First().Value;
            var targetOsisParts = targetOsis.Split('.','-');
            var bookName = _versification.GetBookName(targetOsisParts[0].ToLower());
            string verse = "1";
            if(targetOsisParts.Length >= 3)
            {
               verse = targetOsisParts[2];
            }
            target = string.Format (
               "{0} {1}:{2}", 
               bookName, 
               targetOsisParts[1],
               verse);
         }
         stringBuilder.AppendFormat("<a href=\"{0}\">", target);
         ParseNode ((XElement)elementNode, stringBuilder, verseIdentifier);
         stringBuilder.Append ("</a>");
      }
      
      void ProcessNoteElement(XElement elementNode, StringBuilder stringBuilder, IVerseIdentifier verseIdentifier)
      {
         var numbers = elementNode.Attributes("n");
         if(numbers.Any())
         {
            stringBuilder.AppendFormat("<a href=\"#\" class='notelink'>{0}</a>", numbers.First().Value);
         }
         stringBuilder.Append ("<div class=\"note\">");
         ParseNode ((XElement)elementNode, stringBuilder, verseIdentifier);
         stringBuilder.Append ("</div>");
      }
      
      void ProcessTitleElement(XElement elementNode, StringBuilder stringBuilder, IVerseIdentifier verseIdentifier)
      {
         stringBuilder.Append ("<div id=\"heading\">");
         var originalTextAdded = _textAdded;
         ParseNode ((XElement)elementNode, stringBuilder, verseIdentifier);
         stringBuilder.Append ("</div>");
         stringBuilder.AppendFormat ("<a name=\"{0}\"></a>", verseIdentifier);
         stringBuilder.Append ("<div class=\"verse\">");
         if(IncludeVerseNumbers)
         {
            _verseNumberIndex = stringBuilder.Length;
         }
         _textAdded = originalTextAdded;
         _titleFound = true;
      }
      
      void ProcessQuoteElement(XElement elementNode, StringBuilder stringBuilder, IVerseIdentifier verseIdentifier)
      {
         string quoteMarker = GetQuoteMarker(elementNode);

         if(!elementNode.Attributes("sID").Any() && !elementNode.Attributes("eID").Any())
         {
            stringBuilder.Append(quoteMarker);
         }

         ParseNode (elementNode, stringBuilder, verseIdentifier);
         stringBuilder.Append(quoteMarker);
      }
  
      void ProcessLineGroupElement(XElement elementNode, StringBuilder stringBuilder)
      {
         if(!elementNode.Attributes("eID").Any())
         {
            stringBuilder.Append("<br />");  
         }
      }
      
      void ProcessLineElement(XElement elementNode, StringBuilder stringBuilder, IVerseIdentifier verseIdentifier)
      {
         
         if(!elementNode.Attributes("eID").Any())
         {
            if(_isFirstline && !_textAdded)
            {
               stringBuilder.Append("<div class='firstLine'>");
               if(IncludeVerseNumbers)
               {
                  _verseNumberIndex = stringBuilder.Length;
                  stringBuilder.AppendFormat(" ");
               }
            }
            else
               stringBuilder.Append("<div class='line'>");
            _isFirstline = false;
            
            ParseNode ((XElement)elementNode, stringBuilder, verseIdentifier);
         }
         else
         {
            stringBuilder.Append("</div>");
         }
      }

      void ProcessMilestoneElement(XElement elementNode, StringBuilder stringBuilder)
      {
         if(elementNode.Attributes("type").Any() && elementNode.Attribute("type").Value == "cQuote")
         {
            stringBuilder.Append(GetQuoteMarker(elementNode));
         }
      }

      string GetQuoteMarker(XElement quoteElement)
      {
         var markers = quoteElement.Attributes("marker");
         if(!markers.Any())
         {
            return "\"";
         }
         return (string)markers.First();
      }

      public string ChapterTitle (IVerseIdentifier verseIdentifier)
      {
         return string.Format ("<h1>Chapter {0}</h1>", verseIdentifier.ChapterNumber);
      }
   }
}