// SharpSword C# library for accessing sword modules
// Copyright (C) 2013  Daniel Hughes
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
using System;
using System.Collections.Generic;
using System.Collections;
using System.Linq;
using Lucene.Net.Analysis;
using Lucene.Net.QueryParsers;
using Lucene.Net.Analysis.Standard;
using Lucene.Net.Search;
using Lucene.Net.Store;
using Lucene.Net;
using SharpSword.reader;
using Lucene.Net.Index;
using Lucene.Net.Documents;
using System.IO;

namespace SharpSword
{
   class LuceneIndex : ILuceneIndex
   {
      public static string FIELD_KEY = "key";
      public static string FIELD_BODY = "content";
      public static string FIELD_STRONG = "strong";
      public static string FIELD_HEADING = "heading";
      public static string FIELD_XREF = "xref";
      public static string FIELD_NOTE = "note";
      public static string FIELD_BOOK = "book";
      private string _indexPath;
      private IVersificationInternal _verification;
      private IndexSearcher _searcher;
    
      public LuceneIndex (string modulePath, IVersificationInternal versification)
      {
         _verification = versification;
         _indexPath = Path.Combine (modulePath, "lucenenet");
      }

      private IndexSearcher GetIndexSearcher()
      {
         if(_searcher == null)
         {
            Lucene.Net.Store.Directory directory = new SimpleFSDirectory (new System.IO.DirectoryInfo (_indexPath));
            _searcher = new IndexSearcher (directory);
         }
         return _searcher;
      }

      public IEnumerable<VerseIdentifier> Search (string searchTerm, string book, int start, int numberOfHits)
      {
         int bookNumber = _verification.GetBookNumberFromBookName(book);
         string multiFieldQueryText = string.Format(
            "{0}:{1} AND {2}:{3}",
            FIELD_BODY,
            searchTerm,
            FIELD_BOOK,
            bookNumber);

         return Search(multiFieldQueryText, start, numberOfHits);
      }

      public IDictionary<string, int> GetBookFrequencies(string searchTerm)
      {
         var version = Lucene.Net.Util.Version.LUCENE_30;
         Analyzer analyzer = new StandardAnalyzer (version);

         QueryParser parser = new QueryParser (version, FIELD_BODY, analyzer);
         parser.DefaultOperator = QueryParser.Operator.AND;

         parser.AllowLeadingWildcard = true;
         Query query = parser.Parse (searchTerm);

         IndexSearcher searcher = GetIndexSearcher();

         var bookCountCollector = new BookCountCollector(searcher);
         searcher.Search (query, bookCountCollector);
         var bookHitCounts = bookCountCollector.Frequencies;

         var dictionary = new Dictionary<string, int>();
         for(int bookIndex = 0; bookIndex < 66; bookIndex++)
         {
            int bookCount = (int)bookHitCounts[bookIndex];
            if(bookCount != 0)
            {
               dictionary.Add (_verification.GetBookNameFromBookNumber(bookIndex), bookCount);
            }
         }

         return dictionary;
      }

      public IEnumerable<VerseIdentifier> Search (string searchTerm, int start, int numberOfHits)
      {
         var version = Lucene.Net.Util.Version.LUCENE_30;
         Analyzer analyzer = new StandardAnalyzer (version);

         QueryParser parser = new QueryParser (version, FIELD_BODY, analyzer);
         parser.DefaultOperator = QueryParser.Operator.AND;

         parser.AllowLeadingWildcard = true;
         Query query = parser.Parse (searchTerm);

         IndexSearcher searcher = GetIndexSearcher();

         TopDocs hits = searcher.Search (query, null, start + numberOfHits, Sort.INDEXORDER);

         for (int index = start; index < start + numberOfHits && index < hits.TotalHits; index++)
         {
            var key = searcher.Doc (hits.ScoreDocs [index].Doc).GetField (FIELD_KEY).StringValue;
            yield return VerseIdentifier.ParseFromLuceneKey(key, _verification);
         }
      }
    
      public void GenerateIndex (IBibleModule bibleModule, IVersificationInternal versification)
      {
         using (var directory = FSDirectory.Open (_indexPath))
         using (var analyzer = new StandardAnalyzer (Lucene.Net.Util.Version.LUCENE_29))
         using (var writer = new IndexWriter(directory, analyzer, true, IndexWriter.MaxFieldLength.UNLIMITED))
         {
            var versesQuery =
               from book in versification.GetBookNames ()
               from chapter in versification.GetChapters(book)
               from verse in bibleModule.GetChapter(chapter)
               select verse;

            foreach(Verse verse in versesQuery)
            {
               IVerseIdentifier verseIdentifier = verse.Identifier;
               string osisBookName = versification.GetOsisId(verseIdentifier.BookName);
               string verseContent = verse.Text;
               int bookNumber = _verification.GetBookNumberFromBookName(verseIdentifier.BookName);
               Document doc = new Document ();
               doc.Add (
                  new Field (
                     FIELD_KEY,
                     string.Format ("{0}.{1}.{2}", osisBookName, verseIdentifier.ChapterNumber, verseIdentifier.VerseNumber),
                     Field.Store.YES,
                     Field.Index.NO));
               doc.Add (
                  new Field (
                     FIELD_BOOK,
                     bookNumber.ToString(),
                     Field.Store.YES,
                     Field.Index.ANALYZED));
               doc.Add (
                  new Field (
                     FIELD_BODY,
                     verseContent,
                     Field.Store.NO,
                     Field.Index.ANALYZED));
               writer.AddDocument (doc);
            }

            writer.Optimize ();
         }
      }

      protected void Dispose (bool disposing)
      {
         if(_searcher != null)
         {
            _searcher.Dispose();
            _searcher = null;
         }
      }

      public void Dispose()
      {
         GC.SuppressFinalize(this);
         Dispose(true);
      }
   }
}

