// SharpSword C# library for accessing sword modules
// Copyright (C) 2013  Daniel Hughes
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
using System;
using System.Net;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using ICSharpCode.SharpZipLib.GZip;
using ICSharpCode.SharpZipLib.Tar;
using System.Text;
using LuceneStore = Lucene.Net.Store;
using LuceneUtil = Lucene.Net.Util;
using Lucene.Net;
using Lucene.Net.Analysis;
using Lucene.Net.Analysis.Standard;
using Lucene.Net.Index;
using Lucene.Net.Documents;
using SharpSword.reader;
using System.Threading.Tasks;

namespace SharpSword
{
   public class ModuleManager
   {
      private static readonly string CROSSWIRE_FTP = @"ftp://ftp.crosswire.org/pub/sword/raw/";
      private static readonly string SWORD_APPDATA = GetSwordAppDataPath();
        
      private static readonly string SWORD_MODD = Path.Combine (SWORD_APPDATA, "mods.d");

      private Func<string, IVersificationInternal, ILuceneIndex> _luceneIndexFactory;
      private Func<IFtp> _ftpFactory;
      private Func<string, IModuleConfig> _moduleConfigFactory;
      private IDirectory _directory;
      private Func<string, IBibleModule> _moduleFactory;
      private IFile _file;
    
      public static string GetSwordAppDataPath ()
      {
         if (Environment.OSVersion.Platform == PlatformID.Unix)
         {
            return  Path.Combine (Environment.GetFolderPath (Environment.SpecialFolder.MyDocuments), ".sword");
         }
         return Path.Combine(Environment.GetFolderPath(Environment.SpecialFolder.ApplicationData), "Sword");
      }

      internal ModuleManager (
         Func<string, IVersificationInternal, ILuceneIndex> luceneIndexFactory,
         Func<IFtp> ftpfactory,
         Func<string, IModuleConfig> moduleConfigFactory,
         IDirectory directory,
         Func<string, IBibleModule> moduleFactory,
         IFile file)
      {
         if (!directory.Exists (SWORD_APPDATA))
         {
            directory.CreateDirectory (SWORD_APPDATA);
         }
         if (!directory.Exists (SWORD_MODD))
         {
            directory.CreateDirectory (SWORD_MODD);
         }

         _luceneIndexFactory = luceneIndexFactory;
         _ftpFactory = ftpfactory;
         _moduleConfigFactory = moduleConfigFactory;
         _directory = directory;
         _moduleFactory = moduleFactory;
         _file = file;
      }

      private static ModuleManager _instance;

      public static ModuleManager Instance
      {
         get
         {
            if(_instance == null)
            {
               _instance = new ModuleManager(
                  (modulePath, versification) => new LuceneIndex (modulePath, versification),
                  () => new Ftp(),
                  path => new ModuleConfig(path),
                  new DirectoryWrapper(),
                  path => new BibleZtextReader(path, "en", true, "kjv"),
                  new FileWrapper());
            }
            return _instance;
         }
      }
    
      public IEnumerable<IModuleConfig> GetDownloadableModules ()
      {
         var compressedFileName = "mods.d.tar.gz";
         var path = Path.Combine (CROSSWIRE_FTP, compressedFileName);
         var tempPath = Path.GetTempPath ();
         var pathToSaveTo = Path.Combine (tempPath, compressedFileName);
         var webClient = _ftpFactory();
         webClient.DownloadFile (path, pathToSaveTo);

         _file.ExtractGZip(pathToSaveTo, tempPath);
         var modsdfolder = Path.Combine (tempPath, "mods.d");
       
         return 
            from file in _directory.GetFiles (modsdfolder)
            select _moduleConfigFactory(file);
      }

      private IModuleConfig GetModuleConfig(string moduleName)
      {
         var swordConfigPath = Path.Combine (SWORD_APPDATA, "mods.d", moduleName + ".conf");
         return _moduleConfigFactory(swordConfigPath);
      }

      private string GetModulePath (string moduleName)
      {
         var config = GetModuleConfig(moduleName);
         string dataPath = config.DataPath;
         if (dataPath [0] == '.')
         {
            dataPath = dataPath.Substring (2, dataPath.Length - 2);
            dataPath = Path.Combine(dataPath.Split(new string[]{"/"}, StringSplitOptions.RemoveEmptyEntries));
         }
       
         return Path.Combine (SWORD_APPDATA, dataPath);
      }
    
      string CombineUris (params string[] uriParts)
      {
         var uri = new Uri (uriParts [0]);
         for (int index = 1; index < uriParts.Length; index++)
         {
            uri = new Uri(uri, uriParts[index]);
         }
         return uri.ToString();               
      }
        
      public void InstallModule (string moduleName)
      {
         var ftp = _ftpFactory();
         var crosswireConfigPath = CombineUris(CROSSWIRE_FTP, @"mods.d/", moduleName + ".conf");
         var swordConfigPath = Path.Combine (SWORD_APPDATA, "mods.d", moduleName + ".conf.tmp");

         ftp.DownloadFile (crosswireConfigPath, swordConfigPath);   

         var config = _moduleConfigFactory(swordConfigPath);
       
         string dataPath = config.DataPath;
         if (dataPath [0] == '.')
         {
            dataPath = dataPath.Substring (2, dataPath.Length - 2);
         }
       
         var crosswireModulePath = CombineUris(CROSSWIRE_FTP, dataPath);
         var swordModulePath = Path.Combine (SWORD_APPDATA, Path.Combine(dataPath.Split('/')));

         if (!_directory.Exists (swordModulePath))
         {
            _directory.CreateDirectory (swordModulePath);
         }
         var files = ftp.GetFiles(crosswireModulePath).ToArray();
         Parallel.ForEach(files, ftpFileName =>
         {
            var ftpFilePath = CombineUris(crosswireModulePath, ftpFileName);
            var sharpSwordFilePath = Path.Combine (swordModulePath, ftpFileName);
            bool done = false;
            while(!done)
            {
               try
               {
                  ftp.DownloadFile (ftpFilePath, sharpSwordFilePath);
                  done = true;
               }
               catch(WebException webException)
               {
                  if(webException.Status != WebExceptionStatus.Timeout)
                  {
                     throw;
                  }
               }
            }
         });
         GenerateIndex(config, swordModulePath);
         
         //rename config
         _file.Move(swordConfigPath, swordConfigPath.Remove(swordConfigPath.Length - ".tmp".Length));
      }

      private void GenerateIndex (IModuleConfig config, string swordModulePath)
      {
         IVersificationInternal versification = VersificationManager.Instance.GetVersification (config.Versification);

         using (ILuceneIndex index = _luceneIndexFactory(swordModulePath, versification))
         {

            IBibleModule bibleModule = _moduleFactory (swordModulePath);

            index.GenerateIndex (bibleModule, versification);
         }
      }

      public IEnumerable<IModuleConfig> InstalledModules ()
      {
         return 
            from file in _directory.GetFiles (SWORD_MODD)
            where Path.GetExtension(file) == ".conf"
            select _moduleConfigFactory (file);
      }
    
      public IBibleModule GetModule (string moduleName)
      {
         var config = GetModuleConfig(moduleName);
         var modulePath = GetModulePath (moduleName);
         if(!_directory.Exists(Path.Combine(modulePath, "lucenenet")))
         {
            GenerateIndex(config, modulePath);
         }
         return _moduleFactory(modulePath);
      }
   }
}

