// SharpSword C# library for accessing sword modules
// Copyright (C) 2013  Daniel Hughes
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
using System.Collections.Generic;

namespace SharpSword
{
   /// <summary>
   /// Interface to an object which knows what books chapters and verses exist in a bible.
   /// This provides everything you need to navigate arround a bible module.
   /// </summary>
   public interface IVersification
   {
      /// <summary>
      /// Gets the full names of all the books in biblical order.
      /// </summary>
      /// <returns>Full names of all the books</returns>
      IEnumerable<string> GetBookNames ();

      /// <summary>
      /// Gets the first verse in each chapter in the book.
      /// </summary>
      /// <returns>first verse in each chapter</returns>
      /// <param name='book'>Full name of book</param>
      IEnumerable<IVerseIdentifier> GetChapters (string book);

      /// <summary>
      /// Gets the first verse in the next chapter or null if at the end of the bible.
      /// </summary>
      /// <returns>identifier for first verse in the next chapter or null if we are at the end of the bible</returns>
      /// <param name='verseIdentifier'>identifier for verse in the chapter prior to the one you want.</param>
      IVerseIdentifier NextChapter (IVerseIdentifier verseIdentifier);
    
      /// <summary>
      /// Gets the first verse in the next chapter or null if at the beginning of the bible.
      /// </summary>
      /// <returns>identifier for first verse in the previous chapter or null if we are at the beginning of the bible</returns>
      /// <param name='verseIdentifier'>identifier for verse in the chapter after to the one you want.</param>
      IVerseIdentifier PreviousChapter (IVerseIdentifier verseIdentifier);

      /// <summary>
      /// Determines whether the specified book, chapter and verse represents a valid verse
      /// </summary>
      /// <param name='book'>full name of the book</param>
      /// <param name='chapter'>chapter number</param>
      /// <param name='verse'>verse number</param>
      /// <returns>true if valid else false</returns>
      bool IsValidVerse (string book, int chapter, int verse);

      /// <summary>
      /// Tries to create the create verse identifier.
      /// </summary>
      /// <returns>true if the verse is valid otherwise false.</returns>
      /// <param name='book'>full name of book</param>
      /// <param name='chapter'>chapter number</param>
      /// <param name='verse'>verse number</param>
      /// <param name='verseIdentifier'>VerseIdentifier representing the verse, if the verse if valid</param>
      bool TryCreateVerseIdentifier (string book, int chapter, int verse, out IVerseIdentifier verseIdentifier);

      /// <summary>
      /// Create a verse identifier.
      /// </summary>
      /// <param name='book'>Full book name</param>
      /// <param name='chapter'>Chapter number</param>
      /// <param name='verse'>verse.</param>
      /// <returns>The verse identifier.</returns>
      /// <exception cref="ArgumentOutOfRangeException">
      /// thrown if the combination of book, chapter number and verse number does not represent a valid verse.
      /// </exception>
      IVerseIdentifier CreateVerseIdentifier (string book, int chapter, int verse);

      /// <summary>
      /// Tries to create the create verse identifier.
      /// </summary>
      /// <returns>true if the verse is valid otherwise false.</returns>
      /// <param name='verse'>verse in the standard format e.g Genesis 1:3</param>
      /// <param name='verseIdentifier'>VerseIdentifier representing the verse, if the verse if valid</param>
      bool TryCreateVerseIdentifier (string verse, out IVerseIdentifier verseIdentifier);

      /// <summary>
      /// Create a verse identifier.
      /// </summary>
      /// <param name='verse'>verse in the standard format e.g Genesis 1:3</param>
      /// <exception cref="ArgumentOutOfRangeException">
      /// thrown if the combination the verse string does not represent a valid verse.
      /// </exception>
      IVerseIdentifier CreateVerseIdentifier (string verse);
   }
}