// SharpSword C# library for accessing sword modules
// Copyright (C) 2013  Daniel Hughes
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
using System;
using System.Collections.Generic;
using System.Linq;

namespace SharpSword
{
   class KjvVersification : IVersificationInternal
   {
      Dictionary<string, int> _bookNamesToAbsoluteChapters;
      Dictionary<string, string> _osisIdToBookNames;
      Dictionary<string, string> _bookNamesToOsis;
      Dictionary<int, string> _absoluteChapterNumbersToBookNames;
      Dictionary<string, int> _bookNamesToBookNumber;
      List<string> _books;
    
      internal KjvVersification ()
      {
         _bookNamesToAbsoluteChapters = new Dictionary<string, int> ();
         _osisIdToBookNames = new Dictionary<string, string> ();
         _bookNamesToOsis = new Dictionary<string, string> ();
         _absoluteChapterNumbersToBookNames = new Dictionary<int, string> ();
         _books = new List<string> ();
         _bookNamesToBookNumber = new Dictionary<string, int> ();
         AddBooks ();
      }

      public int GetBookNumberFromBookName(string bookName)
      {
         return _bookNamesToBookNumber[bookName.ToLower()];
      }
    
      public bool TryCreateVerseIdentifier (string book, int chapter, int verse, out IVerseIdentifier verseIdentifier)
      {
         if (!IsValidVerse (book, chapter, verse))
         {
            verseIdentifier = null;
            return false;
         }
         verseIdentifier = new VerseIdentifier (CapitalizeBook(book), chapter, verse);
         return true;
      }
      
      string CapitalizeBook(string book)
      {
         return _books[_bookNamesToBookNumber[book.ToLower()]];
      }
    
      public IVerseIdentifier CreateVerseIdentifier (string book, int chapter, int verse)
      {
         if (!IsValidVerse (book, chapter, verse))
         {
            string message = "book {0}, chapter {1} and verse {2}";
            throw new ArgumentOutOfRangeException (string.Format (message, book, chapter, verse));
         }
         return new VerseIdentifier (CapitalizeBook(book), chapter, verse);
      }

      public bool TryCreateVerseIdentifier (string verse, out IVerseIdentifier verseIdentifier)
      {
         verseIdentifier = null;
         verse = verse.TrimEnd(' ',':');
         if(!verse.Contains (" "))
         {
            return TryCreateVerseIdentifier(verse, 1, 1, out verseIdentifier);
         }
   
         int chapter;
         string book;
         string[] parts;
         
         if(!verse.Contains (":"))
         {
            if(_bookNamesToOsis.ContainsKey(verse.ToLower()))
            {
               //this is a book only identifier with a book number e.g '2 John'
               return TryCreateVerseIdentifier(verse, 1, 1, out verseIdentifier);
            }
            int indexOfLastSpace = verse.LastIndexOf(' ');
            book = verse.Substring(0, indexOfLastSpace);
            if(!int.TryParse(verse.Split(' ').Last(), out chapter))
            {
               return false;
            }
            
            return TryCreateVerseIdentifier(book, chapter, 1, out verseIdentifier);
         }

         parts = verse.Split(':');
         int verseNumber;
         if(!int.TryParse(parts[1], out verseNumber))
         {
            return false;
         }
         string chapterBook = parts[0];
         if(!int.TryParse(chapterBook.Split(' ').Last(), out chapter))
         {
            return false;
         }
         if(!chapterBook.Contains (' '))
         {
            return false;  
         }
         book = chapterBook.Substring(0, chapterBook.LastIndexOf(' '));

         return TryCreateVerseIdentifier(book,chapter, verseNumber, out verseIdentifier);
      }

      public IVerseIdentifier CreateVerseIdentifier (string verse)
      {
         IVerseIdentifier verseIdentifier;

         if(!TryCreateVerseIdentifier(verse, out verseIdentifier))
         {
            throw new ArgumentOutOfRangeException("verse must be in the format 'book chapter:verse' and represent a valid verse");
         }

         return verseIdentifier;
      }
    
      private void AddBooks ()
      {
         AddBook ("Genesis", "gen", 0);
         AddBook ("Exodus", "exod", 50);
         AddBook ("Leviticus", "lev", 90);
         AddBook ("Numbers", "num", 117);
         AddBook ("Deuteronomy", "deut", 153);
         AddBook ("Joshua", "josh", 187);
         AddBook ("Judges", "judg", 211);
         AddBook ("Ruth", "ruth", 232);
         AddBook ("1 Samuel", "1sam", 236);
         AddBook ("2 Samuel", "2sam", 267);
         AddBook ("1 Kings", "1kgs", 291);
         AddBook ("2 Kings", "2kgs", 313);
         AddBook ("1 Chronicles", "1chr", 338);
         AddBook ("2 Chronicles", "2chr", 367);
         AddBook ("Ezra", "ezra", 403);
         AddBook ("Nehemiah", "neh", 413);
         AddBook ("Esther", "esth", 426);
         AddBook ("Job", "job", 436);
         AddBook ("Psalms", "ps", 478);
         AddBook ("Proverbs", "prov", 628);
         AddBook ("Ecclesiastes", "eccl", 659);
         AddBook ("Song of Solomon", "song", 671);
         AddBook ("Isaiah", "isa", 679);
         AddBook ("Jeremiah", "jer", 745);
         AddBook ("Lamentations", "lam", 797);
         AddBook ("Ezekiel", "ezek", 802);
         AddBook ("Daniel", "dan", 850);
         AddBook ("Hosea", "hos", 862);
         AddBook ("Joel", "joel", 876);
         AddBook ("Amos", "amos", 879);
         AddBook ("Obadiah", "obad", 888);
         AddBook ("Jonah", "jonah", 889);
         AddBook ("Micah", "mic", 893);
         AddBook ("Nahum", "nah", 900);
         AddBook ("Habakkuk", "hab", 903);
         AddBook ("Zephaniah", "zeph", 906);
         AddBook ("Haggai", "hag", 909);
         AddBook ("Zechariah", "zech", 911);
         AddBook ("Malachi", "mal", 925);
         AddBook ("Matthew", "matt", 929);
         AddBook ("Mark", "mark", 957);
         AddBook ("Luke", "luke", 973);
         AddBook ("John", "john", 997);
         AddBook ("Acts", "acts", 1018);
         AddBook ("Romans", "rom", 1046);
         AddBook ("1 Corinthians", "1cor", 1062);
         AddBook ("2 Corinthians", "2cor", 1078);
         AddBook ("Galatians", "gal", 1091);
         AddBook ("Ephesians", "eph", 1097);
         AddBook ("Philippians", "phil", 1103);
         AddBook ("Colossians", "col", 1107);
         AddBook ("1 Thessalonians", "1thess", 1111);
         AddBook ("2 Thessalonians", "2thess", 1116);
         AddBook ("1 Timothy", "1tim", 1119);
         AddBook ("2 Timothy", "2tim", 1125);
         AddBook ("Titus", "titus", 1129);
         AddBook ("Philemon", "phlm", 1132);
         AddBook ("Hebrews", "heb", 1133);
         AddBook ("James", "jas", 1146);
         AddBook ("1 Peter", "1pet", 1151);
         AddBook ("2 Peter", "2pet", 1156);
         AddBook ("1 John", "1john", 1159);
         AddBook ("2 John", "2john", 1164);
         AddBook ("3 John", "3john", 1165);
         AddBook ("Jude", "jude", 1166);
         AddBook ("Revelation", "rev", 1167);
      }
    
      void AddBook (string bookName, string osisId, int absoluteChapterNumber)
      {
         _bookNamesToAbsoluteChapters.Add (bookName.ToLower(), absoluteChapterNumber);
         _osisIdToBookNames.Add (osisId, bookName);
         _bookNamesToOsis.Add (bookName.ToLower(), osisId);
         _absoluteChapterNumbersToBookNames.Add (absoluteChapterNumber, bookName);
         _books.Add (bookName);
         _bookNamesToBookNumber.Add (bookName.ToLower(), _books.Count - 1);
      }
    
      public IEnumerable<string> GetBookNames ()
      {
         return _books;
      }

      public string GetOsisId (string bookName)
      {
         return _bookNamesToOsis [bookName.ToLower()];
      }

      public string GetBookName (string osisId)
      {
         return _osisIdToBookNames[osisId];
      }

      public int GetAbsoluteChapterNumber (string bookName)
      {
         return _bookNamesToAbsoluteChapters [bookName.ToLower ()];
      }

      public string GetBookName (int absoluteChapterNumber)
      {
         return _absoluteChapterNumbersToBookNames [absoluteChapterNumber];
      }
    
      public int GetAbsoluteChapterNumberFromOsisId (string osisId)
      {
         string bookName = _osisIdToBookNames [osisId];
         return _bookNamesToAbsoluteChapters [bookName.ToLower()];
      }
    
      public bool DoesOsisIdExist (string osisId)
      {
         return _osisIdToBookNames.ContainsKey (osisId);  
      }
    
      public string GetBookNameFromBookNumber (int bookNumber)
      {
         return _books [bookNumber];
      }
    
      public string GetOsisIdFromBookNumber (int bookNumber)
      {
         string bookName = _books [bookNumber];
         return _bookNamesToOsis [bookName.ToLower ()];  
      }
    
      public bool IsValidVerse (string book, int chapter, int verse)
      {
         if (!_bookNamesToAbsoluteChapters.ContainsKey (book.ToLower()))
         {
            return false;
         }
       
         int bookNumber = _bookNamesToBookNumber [book.ToLower()];
         short[] verseCountInChapters = _versesInChapters [bookNumber];
       
         if (chapter > _versesInChapters [bookNumber].Length)
         {
            return false;
         }
       
         if (verse > verseCountInChapters [chapter - 1])
         {
            return false;   
         }
       
         return true;
      }
    
      public short[][] GetVersesInChapters()
      {
         return _versesInChapters;
      }
    
      public IEnumerable<IVerseIdentifier> GetChapters (string book)
      {
         int bookNumber = _bookNamesToBookNumber [book.ToLower ()];
         short[] verseCountInChapters = _versesInChapters [bookNumber];
         return 
            from chapterNumber in Enumerable.Range (1, verseCountInChapters.Length)
            select new VerseIdentifier (_books[bookNumber], chapterNumber, 1);
      }
    
      public IVerseIdentifier NextChapter (IVerseIdentifier verseIdentifier)
      {
         int newChapter = verseIdentifier.ChapterNumber + 1;
         string bookName = verseIdentifier.BookName;
         int bookIndex = _bookNamesToBookNumber [bookName.ToLower()];
         //check if we are at end of book
         if (newChapter >= _versesInChapters [bookIndex].Length)
         {
            newChapter = 1;
            //check if we are at last book
            if (bookIndex >= _books.Count - 1)
            {
               return null;
            }
            bookName = _books [bookIndex + 1];
         }
         return new VerseIdentifier (bookName, newChapter, 1);
      }
    
      public IVerseIdentifier PreviousChapter (IVerseIdentifier verseIdentifier)
      {
         int newChapter = verseIdentifier.ChapterNumber - 1;
         string bookName = verseIdentifier.BookName;
         int bookIndex = _bookNamesToBookNumber [bookName.ToLower()];
         //check if we are at beginning of book
         if (newChapter == 0)
         {
            //check if we are at first book
            if (bookIndex == 0)
            {
               return null;
            }
            newChapter = _versesInChapters [bookIndex - 1].Length;
            bookName = _books [bookIndex - 1];
         }
         return new VerseIdentifier (bookName, newChapter, 1);
      }

      readonly short[][] _versesInChapters = new short[][]
      {
         new short[]
         {
              31, 25, 24, 26, 32, 22, 24, 22, 29, 32, 32, 20,
              18, 24, 21, 16, 27, 33, 38, 18, 34, 24, 20, 67,
              34, 35, 46, 22, 35, 43, 55, 32, 20, 31, 29, 43,
              36, 30, 23, 23, 57, 38, 34, 34, 28, 34, 31, 22,
              33, 26
         },
         new short[]
         {
            22, 25, 22, 31, 23, 30, 25, 32, 35, 29, 10, 51,
            22, 31, 27, 36, 16, 27, 25, 26, 36, 31, 33, 18,
            40, 37, 21, 43, 46, 38, 18, 35, 23, 35, 35, 38,
            29, 31, 43, 38
         },
         new short[]
         {
            17, 16, 17, 35, 19, 30, 38, 36, 24, 20, 47, 8, 59,
            57, 33, 34, 16, 30, 37, 27, 24, 33, 44, 23, 55,
            46, 34
         },
         new short[]
         {
            54, 34, 51, 49, 31, 27, 89, 26, 23, 36, 35, 16,
            33, 45, 41, 50, 13, 32, 22, 29, 35, 41, 30, 25,
            18, 65, 23, 31, 40, 16, 54, 42, 56, 29, 34, 13
         },
         new short[]
         {
            46, 37, 29, 49, 33, 25, 26, 20, 29, 22, 32, 32,
            18, 29, 23, 22, 20, 22, 21, 20, 23, 30, 25, 22,
            19, 19, 26, 68, 29, 20, 30, 52, 29, 12
         },
         new short[]
         {
            18, 24, 17, 24, 15, 27, 26, 35, 27, 43, 23, 24,
            33, 15, 63, 10, 18, 28, 51, 9, 45, 34, 16, 33
         },
         new short[]
         {
            36, 23, 31, 24, 31, 40, 25, 35, 57, 18, 40, 15,
            25, 20, 20, 31, 13, 31, 30, 48, 25
         },
         new short[]
         {
            22, 23, 18, 22
         },
         new short[]
         {
            28, 36, 21, 22, 12, 21, 17, 22, 27, 27, 15, 25,
            23, 52, 35, 23, 58, 30, 24, 42, 15, 23, 29, 22,
            44, 25, 12, 25, 11, 31, 13
         },
         new short[]
         {
            27, 32, 39, 12, 25, 23, 29, 18, 13, 19, 27, 31,
            39, 33, 37, 23, 29, 33, 43, 26, 22, 51, 39, 25
         },
         new short[]
         {
            53, 46, 28, 34, 18, 38, 51, 66, 28, 29, 43, 33,
            34, 31, 34, 34, 24, 46, 21, 43, 29, 53
         },
         new short[]
         {
            18, 25, 27, 44, 27, 33, 20, 29, 37, 36, 21, 21,
            25, 29, 38, 20, 41, 37, 37, 21, 26, 20, 37, 20,
            30
         },
         new short[]
         {
            54, 55, 24, 43, 26, 81, 40, 40, 44, 14, 47, 40,
            14, 17, 29, 43, 27, 17, 19, 8, 30, 19, 32, 31, 31,
            32, 34, 21, 30
         },
         new short[]
         {
            17, 18, 17, 22, 14, 42, 22, 18, 31, 19, 23, 16,
            22, 15, 19, 14, 19, 34, 11, 37, 20, 12, 21, 27,
            28, 23, 9, 27, 36, 27, 21, 33, 25, 33, 27, 23
         },
         new short[]
         {
            11, 70, 13, 24, 17, 22, 28, 36, 15, 44
         },
         new short[]
         {
            11, 20, 32, 23, 19, 19, 73, 18, 38, 39, 36, 47,
            31
         },
         new short[]
         {
            22, 23, 15, 17, 14, 14, 10, 17, 32, 3
         },
         new short[]
         {
            22, 13, 26, 21, 27, 30, 21, 22, 35, 22, 20, 25,
            28, 22, 35, 22, 16, 21, 29, 29, 34, 30, 17, 25, 6,
            14, 23, 28, 25, 31, 40, 22, 33, 37, 16, 33, 24,
            41, 30, 24, 34, 17
         },
         new short[]
         {
            6, 12, 8, 8, 12, 10, 17, 9, 20, 18, 7, 8, 6, 7, 5,
            11, 15, 50, 14, 9, 13, 31, 6, 10, 22, 12, 14, 9,
            11, 12, 24, 11, 22, 22, 28, 12, 40, 22, 13, 17,
            13, 11, 5, 26, 17, 11, 9, 14, 20, 23, 19, 9, 6, 7,
            23, 13, 11, 11, 17, 12, 8, 12, 11, 10, 13, 20,
            7, 35, 36, 5, 24, 20, 28, 23, 10, 12, 20, 72, 13,
            19, 16, 8, 18, 12, 13, 17, 7, 18, 52, 17, 16, 15,
            5, 23, 11, 13, 12, 9, 9, 5, 8, 28, 22, 35, 45, 48,
            43, 13, 31, 7, 10, 10, 9, 8, 18, 19, 2, 29, 176,
            7, 8, 9, 4, 8, 5, 6, 5, 6, 8, 8, 3, 18, 3, 3,
            21, 26, 9, 8, 24, 13, 10, 7, 12, 15, 21, 10, 20,
            14, 9, 6
         },
         new short[]
         {
            33, 22, 35, 27, 23, 35, 27, 36, 18, 32, 31, 28,
            25, 35, 33, 33, 28, 24, 29, 30, 31, 29, 35, 34,
            28, 28, 27, 28, 27, 33, 31
         },
         new short[]
         {
            18, 26, 22, 16, 20, 12, 29, 17, 18, 20, 10, 14
         },
         new short[]
         {
            17, 17, 11, 16, 16, 13, 13, 14
         },
         new short[]
         {
            31, 22, 26, 6, 30, 13, 25, 22, 21, 34, 16, 6, 22,
            32, 9, 14, 14, 7, 25, 6, 17, 25, 18, 23, 12, 21,
            13, 29, 24, 33, 9, 20, 24, 17, 10, 22, 38, 22, 8,
            31, 29, 25, 28, 28, 25, 13, 15, 22, 26, 11, 23,
            15, 12, 17, 13, 12, 21, 14, 21, 22, 11, 12, 19,
            12, 25, 24
         },
         new short[]
         {
            19, 37, 25, 31, 31, 30, 34, 22, 26, 25, 23, 17,
            27, 22, 21, 21, 27, 23, 15, 18, 14, 30, 40, 10,
            38, 24, 22, 17, 32, 24, 40, 44, 26, 22, 19, 32,
            21, 28, 18, 16, 18, 22, 13, 30, 5, 28, 7, 47, 39,
            46, 64, 34
         },
         new short[]
         {
            22, 22, 66, 22, 22
         },
         new short[]
         {
            28, 10, 27, 17, 17, 14, 27, 18, 11, 22, 25, 28,
            23, 23, 8, 63, 24, 32, 14, 49, 32, 31, 49, 27, 17,
            21, 36, 26, 21, 26, 18, 32, 33, 31, 15, 38, 28,
            23, 29, 49, 26, 20, 27, 31, 25, 24, 23, 35
         },
         new short[]
         {
            21, 49, 30, 37, 31, 28, 28, 27, 27, 21, 45, 13
         },
         new short[]
         {
            11, 23, 5, 19, 15, 11, 16, 14, 17, 15, 12, 14, 16,
            9
         },
         new short[]
         {
            20, 32, 21
         },
         new short[]
         {
            15, 16, 15, 13, 27, 14, 17, 14, 15
         },
         new short[]
         {
            21
         },
         new short[]
         {
            17, 10, 10, 11
         },
         new short[]
         {
            16, 13, 12, 13, 15, 16, 20
         },
         new short[] { 15, 13, 19 }, new short[] { 17, 20, 19 },
         new short[] { 18, 15, 20 }, new short[] { 15, 23 },
         new short[]
         {
            21, 13, 10, 14, 11, 15, 14, 23, 17, 12, 17, 14, 9,
            21
         },
         new short[] { 14, 17, 18, 6 },
         new short[]
         {
            25, 23, 17, 25, 48, 34, 29, 34, 38, 42, 30, 50,
            58, 36, 39, 28, 27, 35, 30, 34, 46, 46, 39, 51,
            46, 75, 66, 20
         },
         new short[]
         {
            45, 28, 35, 41, 43, 56, 37, 38, 50, 52, 33, 44,
            37, 72, 47, 20
         },
         new short[]
         {
            80, 52, 38, 44, 39, 49, 50, 56, 62, 42, 54, 59,
            35, 35, 32, 31, 37, 43, 48, 47, 38, 71, 56, 53
         },
         new short[]
         {
            51, 25, 36, 54, 47, 71, 53, 59, 41, 42, 57, 50,
            38, 31, 27, 33, 26, 40, 42, 31, 25
         },
         new short[]
         {
            26, 47, 26, 37, 42, 15, 60, 40, 43, 48, 30, 25,
            52, 28, 41, 40, 34, 28, 41, 38, 40, 30, 35, 27,
            27, 32, 44, 31
         },
         new short[]
         {
            32, 29, 31, 25, 21, 23, 25, 39, 33, 21, 36, 21,
            14, 23, 33, 27
         },
         new short[]
         {
            31, 16, 23, 21, 13, 20, 40, 13, 27, 33, 34, 31,
            13, 40, 58, 24
         },
         new short[]
         {
            24, 17, 18, 18, 21, 18, 16, 24, 15, 18, 33, 21,
            14
         },
         new short[] { 24, 21, 29, 31, 26, 18 },
         new short[] { 23, 22, 21, 32, 33, 24 },
         new short[] { 30, 30, 21, 23 },
         new short[] { 29, 23, 25, 18 },
         new short[] { 10, 20, 13, 18, 28 },
         new short[] { 12, 17, 18 },
         new short[] { 20, 15, 16, 16, 25, 21 },
         new short[] { 18, 26, 17, 22 },
         new short[] { 16, 15, 15 }, new short[] { 25 },
         new short[]
         {
            14, 18, 19, 16, 14, 20, 28, 13, 28, 39, 40, 29,
            25
         },
         new short[] { 27, 26, 18, 17, 20 },
         new short[] { 25, 25, 22, 19, 14 },
         new short[] { 21, 22, 18 },
         new short[] { 10, 29, 24, 21, 21 }, new short[] { 13 },
         new short[] { 14 }, new short[] { 25 },
         new short[]
         {
            20, 29, 22, 11, 14, 17, 17, 13, 21, 11, 19, 17,
            18, 20, 8, 21, 18, 24, 21, 15, 27, 21
         }
      };
   }
}

